/*
 * arff generator class
 */

#include "arff.h"

#include "config.h"

#include <strings.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>

#include <sys/types.h>
#include <sys/time.h>

#include <regex.h>
#include <ctype.h>
#include <fcntl.h>

#ifdef HAVE_SYS_RESOURCE_H
#include <sys/resource.h>
#endif

#ifdef HAVE_INTTYPES_H
#ifndef __STDC_FORMAT_MACROS
#define __STDC_FORMAT_MACROS
#endif
#include <inttypes.h>
#endif

#include <err.h>

#include <openssl/evp.h>
#include <openssl/md5.h>
#include <openssl/sha.h>
#include <tsk/libtsk.h>

#ifdef HAVE_AFFLIB
#include <afflib/afflib.h>
#endif

#include <algorithm>
#include <cstdlib>
#include <vector>
#include <string>
#include <assert.h>
using namespace std;

void arff::add_comment(string comment)
{
    comments.push_back(comment);
}

bool arff::has_attribute(const string &name) const
{
    for(int i=0;i<attributeTypes.size();i++){
	if(attributeNames[i]==name) return true;
    }
    return false;
}

/**
 * If the attribut hasn't already been given a type,
 * add a type.
 */

void arff::add_attribute(string name,int code)
{

    for(int i=0;i<attributeTypes.size();i++){
	if(attributeNames[i]==name){
	    return;			// older types dominate newer ones;
	}
    }
    attributeTypes.push_back(code);
    attributeNames.push_back(name);
    attributeCodes[name] = attributeNames.size()-1;
}

void arff::new_row()
{
    values.push_back(new valueMapT);
}

void arff::add_value(string name,int64_t value)
{
    int code = attributeCodes[name];
    int type = attributeTypes[code];
    time_t valuet = value;

    char buf[64];
    switch(type){
    case DATE:
	strftime(buf,sizeof(buf),"%F %T",gmtime((time_t *)&valuet));
	break;
    default:
	sprintf(buf,"%"PRIuDADDR,value);
	break;
    }
    (*values.back())[code] = buf;
}

void arff::add_valuet(string name,time_t t)
{
    int code = attributeCodes[name];
    char buf[64];
    strftime(buf,sizeof(buf),"%F %T",gmtime(&t));
    (*values.back())[code] = buf;
}

void arff::add_value(string name,const string &value)
{
    // If type doesn't exist, create it and make it a STRING
    if(attributeCodes.find(name)==attributeCodes.end()){
	add_attribute(name,STRING);
    }

    int code = attributeCodes[name];
    (*values.back())[code] = value;
}

static bool has_non_numeric(const string &s)
{
    for(int i=0;i<s.size();i++){
	if(isdigit(s[i])==0 && s[i]!='.' && s[i]!='-' && s[i]!='+') return true; // it has a non-numeric
	if(i>0 && s[i]=='-' ) return true;
	if(i>0 && s[i]=='+' ) return true;
    }
    return false;
}

static bool is_date(const string &s)
{
    if(isdigit(s[0]) && isdigit(s[1]) && isdigit(s[2]) && isdigit(s[3]) && s[4]=='-' &&
       isdigit(s[5]) && isdigit(s[6]) && s[7]=='-' &&
       isdigit(s[8]) && isdigit(s[9])  &&
       s[10]==' ' &&
       isdigit(s[11]) && isdigit(s[12]) && s[13]==':' &&       
       isdigit(s[14]) && isdigit(s[15]) && s[16]==':' &&       
       isdigit(s[17]) && isdigit(s[17]) && s.size()==19) return true;
    return false;
}

int arff::attributeCol(const string &attributeName) const
{
    for(int i=0;i<attributeNames.size();i++){
	if(attributeNames[i]==attributeName) return i;
    }
    return -1;
}

/* Scan the values we have for an attribute and return true if it always missing or numeric. */
bool arff::attributeAlwaysNumeric(const string &attributeName) const
{
    int col = attributeCol(attributeName);
    for(vector<valueMapT *>::const_iterator row = values.begin(); row!= values.end(); row++){
	valueMapT::const_iterator k = (*row)->find(col); // find the row we are looking for
	if(k==(*row)->end()) continue;		   // no value
	const string &val = (*k).second.c_str();	   // no need to copy
	if(has_non_numeric(val)) return false;
    }
    return true;
}

bool arff::attributeAlwaysDate(const string &attributeName) const
{
    int col = attributeCol(attributeName);
    for(vector<valueMapT *>::const_iterator row = values.begin(); row!= values.end(); row++){
	valueMapT::const_iterator k = (*row)->find(col); // find the row we are looking for
	if(k==(*row)->end()) continue;		   // no value
	const string &val = (*k).second.c_str();	   // no need to copy
	if(is_date(val)==false) return false;
    }
    return true;
}

/** return true if a string needs quoting */
static bool needs_quoting(const string &str){
    for(int i=0;i<str.size();i++){
	switch(str[i]){
	case ' ':
	case '{':
	case '}':
	case ',':
	case '\t':
	    return true;
	}
    }
    return false;
}

void arff::write()
{
    /* Before we generate the output, see if any of the string attributes always have numeric values.
     * If so, we'll change them into numeric values...
     */
    for(int i=0;i<attributeNames.size();i++){
	if(attributeTypes[i]==STRING && attributeAlwaysNumeric(attributeNames[i])){
	    attributeTypes[i]=NUMERIC;
	}
	if(attributeTypes[i]==STRING && attributeAlwaysDate(attributeNames[i])){
	    attributeTypes[i]=DATE;
	}
    }


    for(vector<string>::const_iterator i = comments.begin();
	i!= comments.end();
	i++){
	fprintf(outfile,"%% %s\n",i->c_str());
    }
    fprintf(outfile,"%\n");
    fprintf(outfile,"@RELATION %s\n",relation.c_str());
    fprintf(outfile,"\n");
    
    for(int i=0;i<attributeNames.size();i++){
	string name = attributeNames[i];

	/* Turn the spaces into underbars */
	for(string::iterator j = name.begin();j!=name.end();j++){
	    if(*j == ' ') *j = '_';
	}
	switch(attributeTypes[i]){
	case NUMERIC:
	    fprintf(outfile,"@ATTRIBUTE %s NUMERIC\n",name.c_str());
	    break;
	case DATE:
	    fprintf(outfile,"@ATTRIBUTE %s date \"yyyy-MM-dd HH:mm:ss\"\n",name.c_str());
	    break;
	case STRING:
	    fprintf(outfile,"@ATTRIBUTE %s string\n",name.c_str());
	    break;
	default:
	    assert(0);
	}
    }
    fprintf(outfile,"\n@DATA\n\n");
    for(vector<valueMapT *>::const_iterator row = values.begin();
	row!= values.end();
	row++){
	for(int i=0;i<attributeNames.size();i++){
	    valueMapT::const_iterator k = (*row)->find(i); // find the row we are looking for
	    string val("?");		// default value
	    if(k!=(*row)->end()){
		val = (*k).second;
		/* Remove any quotes */
		for(int j=0;j<val.size();j++){
		    if(val[j]=='"') val[j]=' ';
		    if(val[j]=='\'') val[j]=' ';
		}
		/* If there are any spaces quote the result */
		if(needs_quoting(val)){
		    val = "\"" + val + "\"";
		}
	    }
	    if(val.size()==0) val="?";	// put back the ?
	    if(i>0) fprintf(outfile,", ");
	    fprintf(outfile,"%s",val.c_str());
	}
	fprintf(outfile,"\n");
    }
    fflush(outfile);
}

