/*
 * fiwalk.cpp:
 * File and Inode Walk.
 *
 * This application uses SleuthKit to generate a report of all of the files
 * and orphaned inodes found in a disk image. It can optionally compute the MD5 of 
 * any objects, save those objects into a directory, or both.
 *
 * Algorithm:
 * 1 - Find all of the partitions on the disk. 
 * 2 - For each partition, walk the files.
 * 3 - For each file, print the requested information.
 * 4 - For each partition, walk the indoes
 * 5 - For each inode, print the requested information.
 */


#include "config.h"

#include <strings.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>

#include <sys/types.h>
#include <sys/time.h>

#include <regex.h>
#include <ctype.h>
#include <fcntl.h>

#ifdef HAVE_SYS_RESOURCE_H
#include <sys/resource.h>
#endif

#ifdef HAVE_INTTYPES_H
#ifndef __STDC_FORMAT_MACROS
#define __STDC_FORMAT_MACROS
#endif
#include <inttypes.h>
#endif

#include <err.h>

#include <openssl/evp.h>
#include <openssl/md5.h>
#include <openssl/sha.h>
#include <tsk/libtsk.h>

#ifdef HAVE_AFFLIB
#include <afflib/afflib.h>
#endif

#include <algorithm>
#include <cstdlib>
#include <vector>
#include <string>
using namespace std;


/* Bring in our headers */
#include "arff.h"
#include "plugin.h"
#include "bloom.h"
#include "bloomset.h"

#include "threaded_hash.h"
#include "tagset.h"
#include "xml.h"


/* Structure for keeping track of file segments */
class seg {
public:;
    int64_t start;			// byte offset from start of image
    int64_t len;				// number of bytes
};

/* Configurable options */
char *fiwalk_version = "0.4";
char *save_outdir = "";
int  opt_debug    = 0;
int  opt_md5      = 0;			// calculate the MD5 for every file?
int  opt_sha1     = 0;			// calculate the SHA1 for every file?
bool opt_save     = false;
bool opt_userguess = false;
bool opt_magic = false;			// get the output of the file command
int  opt_multithreaded = 1;		// use multiple threading
bool do_plugin = false;		// this one requires plugin processing
char *save_bloom512 = 0;
const char *config_file = 0;
int  file_count_max = 0;			// 
int  file_count = 0;

bloomset mybloomset;

typedef vector<class seg> seglist;	// vector of blocks

vector<string> namelist;		// names of files that we want to find

/****************************************************************
 ** Current information.
 **
 ****************************************************************/
int      current_partition_num=0;
int64_t  current_partition_start=0;	// in bytes

/* Individual 'state' variables */
string  plugin_filename;

void print_version()
{
    printf("FIWalk Version:    %s\n",PACKAGE_VERSION);
    printf("SleuthKit Version: %s\n",tsk_version_get_str());
#ifdef HAVE_AFFLIB
    printf("AFFLIB Version:    %s\n",af_version());
#else
    printf("*** NO AFFLIB SUPPORT ***\n");
#endif    
    
}

void usage()
{
    printf("usage: fiwalk [options] iso-name\n");
    printf("Default behavior: Just print the file system statistics and exit.\n");
    printf("options:\n");
    printf("    -c config.txt   read config.txt for metadata extraction tools\n");
    printf("    -C nn           only process nn files, then do a clean exit\n"); 
    printf("    -u              Guess at usernames based on filenames, mode, or UIDs\n");
    printf("    -S              Single threaded; do not use multi-threaded.\n");
    
    printf("\n");
    printf("include/exclude parameters; may be repeated. \n");
    printf("NOTE: DO NOT RELY ON THIS SYNTAX; IT WILL CHANGE (soon)\n");

    printf("    -b+bloomfile.bloom = Only include files whose SHA1 is in bloomfile.bloom\n");
    printf("    -b-bloomfile.bloom = Only include files whose SHA1 is not in bloomfile.bloom\n");
    printf("    -n pattern  = only match files for which the filename matches\n");
    printf("                  the pattern.\n");
    printf("              example: -n .jpeg -n .jpg will find all JPEG files\n");
    printf("              Case is ignored. Will not match orphan files.\n");
    printf("    ");
    printf("\n");
    printf("\nFull content options (using any of these slow processing significantly):\n");
    printf("    -m = Report MD5 for each file\n");
    printf("    -1 = Report SHA1 for each file\n");
    printf("    -s <dir> = Save all recovered files to <dir>\n");
    printf("    -f = Report the output of the 'file' command for each\n");
    printf("\n");
    printf("Output options:\n");
    printf("    -A<file> = ARFF output to <file>\n");
    printf("    -X<file> = XML output to a <file>\n");
    printf("    -T<file> = Walkfile output to <file>\n");
    printf("    -B = output 512-byte bloom filter\n");
    printf("\n");
    printf("    -d = debug this program\n");
    printf("    -v = Enable SleuthKit verbose flag\n");
    printf("\n");
    print_version();
    exit(1);
}

/****************************************************************
 ** Support routines 
 ****************************************************************/
static const char *cstr(const string &str){
    return str.c_str();
}

static string empty("");


/****************************************************************
 ** XML output
 ****************************************************************/

/****************************************************************
 ** Metadata Output
 ****************************************************************/

/* Specify the output mode */
class arff *a = 0;			// ARFF generator
class xml  *x = 0;
FILE  *t = 0;				// text output

/**
 * output a comment in the current file format
 */

void comment(const char *format,...)
{
    char buf[1024];
    va_list ap;
    va_start(ap, format);
    vsnprintf(buf,sizeof(buf),format,ap);
    va_end(ap);

    if(t){
	fputs("# ",t);
	fputs(buf,t);
	fputc('\n',t);
    }
	
    if(x){
	fputs("<!-- ",x->file());
	fputs(buf,x->file());
	fputs(" -->\n",x->file());
    }

    if(a) {
	a->add_comment(buf);
    }
}

/**
 * output a name/format/value for the current partition.
 * This information is simply printed as comments for ARFF files.
 *
 * @param name - the name of the thing being output
 * @param format - the format for the thing
 * @param ... - the value
 * This will output as a comment in the ARFF file
 */
void partition_info(const string &name,const string &value)
{
    if(name.find(" ")<0) err(1,"partition_info(%s) has a space in it",cstr(name));
    if(a) a->add_comment(name + ": " + value);
    if(t) fputs(cstr(name + ": " + value + "\n"),t);
    if(x) x->xmlout(name,value,empty);
}

void partition_info(const string &name,long i)
{
    char buf[1024];
    sprintf(buf,"%d",i);
    partition_info(name,buf);
}

/****************************************************************/
void file_info(const string &name,const string &value)
{
    if(a) a->add_value(name,value); 
    if(t) fputs(cstr(name + ": " + value + "\n"),t); 
    if(x) x->xmlout(name,value,empty); 
}

void file_info(const string name, long long value)
{
    if(a) a->add_value(name,value); 
    if(t || x){
	char buf[1024];
	snprintf(buf,sizeof(buf),"%"PRIuDADDR,value);

	if(t) fputs(cstr(name + ": " + buf + "\n"),t);
	if(x) x->xmlout(name,buf,empty);
    }
}

void file_infot(const string name,time_t t0)
{
    if(a) a->add_valuet(name,t0); 
    if(x){
	char buf[16];
	snprintf(buf,sizeof(buf),"%d",t0);
	x->xmlout(name,buf,empty);
    }
    if(t) {
	char buf[64];
	fprintf(t,"%s: %d\n",name.c_str(),t0);
	strftime(buf,sizeof(buf),"%F %T",gmtime(&t0));
	fprintf(t,"%s_txt: %s\n",name.c_str(),buf);
    }
}

/****************************************************************/
string mytime()
{
    time_t t = time(0);
    char *buf = ctime(&t);
    buf[24] = 0;
    return string(buf);
}

/****************************************************************
 ** Content Output
 ****************************************************************/

/**
 * The content_info class is a class that deals with content when it is discovered.
 */

class content_info {
public:
    string   evidence_dirname;		// where it is being put
    string   evidence_filename;         // filename of what's currently being saved (from evidence file system)

    FILE    *savefile;		        // saving the copy for posterity
    string   save_path;	                // full path of where it is being saved

    FILE    *tempfile;			// copy for plugin and "file" command
    string  tempdir;			// directory where temporary files are put
    string  tempfile_path;		// where the tempfile was put

    threaded_hash h_md5;
    threaded_hash h_sha1;

    seglist segs;			// the segments that make up the file

    content_info():savefile(0),
		   tempfile(0),
		   tempdir("/tmp"),
		   h_md5(EVP_md5,opt_multithreaded),
		   h_sha1(EVP_sha1,opt_multithreaded) {
    }
    bool   has_filename() { return evidence_filename.size()>0;}
    string filename() { return evidence_dirname + evidence_filename;}
    string filemagic();			// returns output of the 'file' command or libmagic
    void open_savefile();
    void open_tempfile();
    void add_bytes(const char *buf,size_t size){
	if(opt_md5)   h_md5.update((unsigned char *)buf,size);
	if(opt_sha1)  h_sha1.update((unsigned char *)buf,size);
	if(savefile)  fwrite(buf,size,1,savefile);
	if(tempfile)  fwrite(buf,size,1,tempfile);
    }
    void final(){
	if(savefile){ fclose(savefile); savefile = 0; }
	if(tempfile){ fclose(tempfile); tempfile = 0; }
    }
    ~content_info(){
	if(tempfile_path.size()>0) unlink(tempfile_path.c_str());
    }
};

/** Open a temporary file to put the file for the plugin and/or the file command
 */
void content_info::open_tempfile()
{
    if(tempfile) return;		//  already opened

    int added = 0;
    tempfile_path = tempdir + "/";
    for(string::const_iterator cc = evidence_filename.begin();
	cc != evidence_filename.end();
	cc++){
	if(isalnum(*cc)) tempfile_path.push_back(*cc);
	added++;
    }
    if(added==0) tempfile_path += "tempfile";

    tempfile = fopen(tempfile_path.c_str(),"w");
    if(!tempfile) errx(1,tempfile_path.c_str());
}


/** Open a file where a bytestream will be saved.
 *
 * @param filename - specifies the name of the file. If the file
 * exists, then start a counter between the end of the file and the
 * extension and keep incrementing...
 *
 */
void content_info::open_savefile()
{
    /* Figure out the pathname to save */
    save_path = save_outdir + string("/") + evidence_filename;

    /* IF we can open this directory directly, just take it */
    if(access(save_path.c_str(),F_OK)!=0){
	savefile = fopen(save_path.c_str(),"w");
	return;
    }

    /* Make a format string that is filename.%d.ext
     */

    string fmtstring = save_path;

    size_t ext = fmtstring.rfind('.');
    if(ext==-1){
	fmtstring += ".%03d";
    }
    else{
	fmtstring.insert(ext,".%03d.");
    }

    for(int i=1;i<10000;i++){
	char fname[1024];
	snprintf(fname,sizeof(fname),fmtstring.c_str(),i);
	if(access(fname,F_OK)!=0){
	    savefile = fopen(fname,"w");
	    return;
	}
    }
    // hm. Can't open file?
}

/**
 * run the file command.
 * -b = do not put the filename in the output.
 * -z = attempt to decompress compressed files.
 */
 
string content_info::filemagic()
{
    char cmd[1024];
    char buf[1024];
    string out;
    snprintf(cmd,sizeof(cmd),"file -b -z '%s'",tempfile_path.c_str());
    FILE *f = popen(cmd,"r");
    while(!feof(f)){
	if(fgets(buf,sizeof(buf),f)) out += buf;
    }
    pclose(f);
    while(out.size()>0 && out[out.size()-1]=='\n'){
	out.erase(out.size()-1);
    }
    return out;
}


/****************************************************************
 *** Support Code.
 ****************************************************************/

/* FS_INFO defined in ~/sleuthkit-2.05/src/fstools/fs_tools.h
 * IMG_INFO defined in img_tools.h 
 * img_open - opens the image; calls do_dimage
 * do_dimage - analyizes the image
 *        mm_open - opens the volume; mm_part_walk - walk each volume;
 *                  calls mm_act for each
 * mm_act - analyizes each partition; calls do_vol for ones that we can analyze
 * do_vol - analyizes each volume
 *       fs_open - opens the file system; dent_walk 
 *               - walks the directory structure.
 *               - calls dent_act for each file
 * dent_act - prints file name; calls file_walk() to read each file; 
 *          - calls file_act for each sector
 * file_act - prints the block number
 */

/** 
 * file_act:
 *  -- obtain the address of each data unit
 *  -- optionally calculate the mac or save the data
 * 
 */
TSK_WALK_RET_ENUM
file_act(TSK_FS_FILE * fs_file, TSK_OFF_T a_off, TSK_DADDR_T addr, char *buf,
	 size_t size, TSK_FS_BLOCK_FLAG_ENUM flags, void *ptr)
{
    content_info *ci = (content_info *)ptr;

    if(opt_debug){
	printf("file_act(fs_file=%p,addr=%"PRIuDADDR" buf=%p size=%d)\n",fs_file,addr,buf,(int)size);
	if(opt_debug>1 && ci->segs.size()==0){
	    fwrite(buf,size,1,stdout);
	    printf("\n");
	}
    }

    if(size==0)  return TSK_WALK_CONT;	    // can't do much with this...
    ci->add_bytes(buf,size);

    /* See if we can extend the last segment in the segment list,
     * or if this is the start of a new fragment.
     */
    int64_t byteaddr = current_partition_start + ((int64_t)addr)*fs_file->fs_info->block_size;

    if(ci->segs.size()>0){
	if(ci->segs.back().start + ci->segs.back().len == (int64_t)byteaddr){
	    ci->segs.back().len += size;
	    return TSK_WALK_CONT;
	}
    }
    /* Need to add a new element to the list */
    struct seg newseg;
    newseg.start = byteaddr;
    newseg.len   = size;
    ci->segs.push_back(newseg);
    return TSK_WALK_CONT;
}


/**
 * Try to guess the username from the path and, if found, add it.
 */
void add_userguess(const string &path)
{
    static bool first = true;
    static regex_t re;
    if(first){
	if(regcomp(&re,"Documents and Settings/([^/]+)/",REG_EXTENDED|REG_ICASE)) err(1,"regcomp");
	first = 0;
    }
    regmatch_t pmatch[4];
    memset(pmatch,0,sizeof(pmatch));
    if(regexec(&re,path.c_str(),2,pmatch,0)) return;
    string username = path.substr(pmatch[1].rm_so,pmatch[1].rm_eo-pmatch[1].rm_so);
    file_info("userguess",username);
}


/* Do we need full content? */
bool full_content()
{
    return opt_md5 || opt_sha1 || opt_save || do_plugin  || (mybloomset.size()>0) || opt_magic; 
}

/**
 * Process the contents of a file.
 *
 * @return 1 on error and 0 on success 
 */
static uint8_t
proc_file(TSK_FS_FILE * fs_file, const char *path)
{
    if ((fs_file->meta == NULL) || (fs_file->name == NULL)) return 1;

    if (fs_file->meta->type != TSK_FS_META_TYPE_REG) return 0;

    if(opt_debug) printf("Processing %s%s\n", path, fs_file->name->name);

    /* Recover the filename from the fs_dent, if it is provided */
    content_info ci;			// where the content will go
    ci.evidence_dirname  = path;
    ci.evidence_filename = fs_file->name->name;

    /* If we are filtering and we have a filename, see if we want this file. */
    if(namelist.size()>0 && ci.evidence_filename.size()>0){
	bool file_wanted = false;
	for(vector<string>::const_iterator i = namelist.begin(); i!=namelist.end(); i++){
	    if(strcasestr(ci.evidence_filename.c_str(),i->c_str())){
		file_wanted = true;
	    }
	}
	if(!file_wanted) return 0;
    }

    /* Looks like we are processing */
    if(a) a->new_row();
    if(x) x->push("fileobject"); 
	    
    /* See if this one requires plugin processing */
    do_plugin = plugin_match(ci.evidence_filename.c_str());
    if(do_plugin || opt_magic)	ci.open_tempfile();
    if(opt_save)	ci.open_savefile();


    if(full_content()){
	int myflags = TSK_FS_FILE_WALK_FLAG_NOID;
	if (tsk_fs_file_walk (fs_file, (TSK_FS_FILE_WALK_FLAG_ENUM) myflags, file_act, (void *) &ci)) {
	    // ignore errors from deleted files that were being recovered
	    if (tsk_errno != TSK_ERR_FS_RECOVER) {
		printf("Processing: %s/%s (%" PRIuINUM ")\n", path,
		       fs_file->name->name, fs_file->meta->addr);
		tsk_error_print(stderr);
	    }
	    tsk_error_reset();
	}
	else {
	    ci.final();
	}
    }

    /* See if the bloomset says to get rid of the file */
    /* This will screw up ARFF and XML; need a way of discarding current object. */
    if(mybloomset.size()>0){
	if(mybloomset.check_exclude(ci.h_sha1.hash())){
	    return TSK_WALK_CONT;
	}
    }

    if(file_count_max && file_count>file_count_max) exit(1);
    file_count++;

    /* Send through to the plugin if we were doing that.
     * Currently results only go to ARFF file, not to the XML file.
     */
    if(do_plugin) plugin_process(ci.tempfile_path.c_str());

    /* Finally output the informaton */
    if(ci.has_filename())			file_info(string("filename"),ci.filename());
    file_info("filesize",fs_file->meta->size);
    if(fs_file->meta->flags & TSK_FS_META_FLAG_ALLOC)   file_info("ALLOC",1);
    if(fs_file->meta->flags & TSK_FS_META_FLAG_UNALLOC) file_info("UNALLOC",1);
    if(fs_file->meta->flags & TSK_FS_META_FLAG_USED)    file_info("USED",1);
    if(fs_file->meta->flags & TSK_FS_META_FLAG_UNUSED)  file_info("UNUSED",1);
    if(fs_file->meta->flags & TSK_FS_META_FLAG_ORPHAN)  file_info("ORPHAN",1);
    if(fs_file->meta->mtime) file_infot("mtime",fs_file->meta->mtime);
    if(fs_file->meta->ctime) file_infot("ctime",fs_file->meta->ctime);
    if(fs_file->meta->atime) file_infot("atime",fs_file->meta->atime);
    if(opt_magic)  file_info("libmagic",ci.filemagic());
    if(ci.segs.size()>0){
	string runz = "";
	for(seglist::const_iterator i = ci.segs.begin();i!=ci.segs.end();i++){
	    char buf[1024];
	    sprintf(buf,"%qd:%qd",i->start,i->len);
	    if(runz.size()>0) runz += "|";
	    runz += buf;
	}
	file_info("byte runs",runz);
	if(opt_md5) file_info("md5",ci.h_md5.hexhash());
	if(opt_sha1) file_info("sha1",ci.h_sha1.hexhash());
    }

    file_info("fragments",ci.segs.size());
    
    file_info("partition",current_partition_num);
    int block_size = fs_file->fs_info->block_size;
    if(block_size>0){
	if(ci.segs.size()>=1) file_info("frag1startsector",(int)(ci.segs[0].start / block_size));
	if(ci.segs.size()>=2) file_info("frag2startsector",(int)(ci.segs[1].start / block_size));
    }
    if(opt_userguess && ci.has_filename()) add_userguess(ci.filename());

    /* END of file processing */
    if(x) x->pop("fileobject");
    if(t) fputs("\n",t);
    return TSK_WALK_CONT;
}




/**
 * file name walk callback.  Walk the contents of each file 
 * that is found.
 */
static TSK_WALK_RET_ENUM
dir_act(TSK_FS_FILE * fs_file, const char *path, void *ptr)
{
    /* Ignore NTFS System files */
    if ((TSK_FS_TYPE_ISNTFS(fs_file->fs_info->ftype))
        && (fs_file->name->name[0] == '$'))
        return TSK_WALK_CONT;

    /* If the name has corresponding metadata, then walk it */
    if (fs_file->meta) {
        proc_file(fs_file, path);
    }

    return TSK_WALK_CONT;
}



int proc_fs(TSK_IMG_INFO * img_info, TSK_OFF_T start)
{
    TSK_FS_INFO *fs_info;

    comment("fs start: %" PRIuDADDR, start);
    if(x) x->printf("<volume startsector='%"PRIuDADDR"'>\n",start);

    /* Try it as a file system */
    if ((fs_info =
            tsk_fs_open_img(img_info, start, TSK_FS_TYPE_DETECT)) == NULL) {
	partition_info("TSK_Error",tsk_error_get());
        /* We could do some carving on the volume data at this point */
	if(x) x->pop("volume");
        return 1;
    }

    current_partition_num++;
    current_partition_start = fs_info->offset;
    partition_info("Partition_Offset",fs_info->offset);
    partition_info("block_size",fs_info->block_size);
    partition_info("ftype",fs_info->ftype);
    partition_info("ftype_str",tsk_fs_type_toname(fs_info->ftype));
    partition_info("block_count",fs_info->block_count);
    partition_info("first_block",fs_info->first_block);
    partition_info("last_block",fs_info->last_block);
    if(t) fputc('\n',t);

    int ret = 0;
    /* Walk the files, starting at the root directory */
    if (tsk_fs_dir_walk(fs_info, fs_info->root_inum,
            (TSK_FS_DIR_WALK_FLAG_ENUM) (TSK_FS_DIR_WALK_FLAG_RECURSE),
            dir_act, NULL)) {
	comment("TSK Error: tsk_fs_dir_walk: ",tsk_error_get());
	ret = 1;
        return 1;
    }

    /* We could do some analysis of unallocated blocks at this point...  */

    tsk_fs_close(fs_info);
    if(x) x->pop("volume");
    comment("end of volume");
    return ret;
}

/**
 * Volume system walk callback function that will analyze 
 * each volume to find a file system.
 */
static TSK_WALK_RET_ENUM
vs_act(TSK_VS_INFO * vs_info, const TSK_VS_PART_INFO * vs_part,
    void *ptr)
{
    if (proc_fs(vs_info->img_info, vs_part->start * vs_info->block_size)) {
        // if we return ERROR here, then the walk will stop.  But, the 
        // error could just be because we looked into an unallocated volume.
        // do any special error handling / reporting here.
        tsk_error_reset();
        return TSK_WALK_CONT;
    }

    return TSK_WALK_CONT;
}


/* do_dimage:
 * Process the partition tables in the disk image pointed to by img.
 * File system analysis will be performed on each partition.
 *   mm_open() - opens partition
 *   mm_part_walk() - walks partition
 * If we can't open as a disk image, try to open as a raw file system
 *
 * returns 1 on error and 0 on success
 */
int proc_vs(TSK_IMG_INFO * img_info, TSK_OFF_T start)
{
    TSK_VS_INFO *vs_info;

    // USE mm_walk to get the volumes 
    if ((vs_info =
            tsk_vs_open(img_info, start, TSK_VS_TYPE_DETECT)) == NULL) {
        if (tsk_verbose)
            fprintf(stderr,
                "Error determining volume system -- trying file systems\n");

        /* There was no volume system, but there could be a file system */
        tsk_error_reset();
        if (proc_fs(img_info, start)) {
            return 1;
        }
    }
    else {
        if (tsk_verbose) fprintf(stderr, "Volume system open, examining each\n");

        /* Walk the allocated volumes (skip metadata and unallocated volumes) */
        if (tsk_vs_part_walk(vs_info, 0, vs_info->part_count-1,
                (TSK_VS_PART_FLAG_ENUM) (TSK_VS_PART_FLAG_ALLOC), vs_act, NULL)) {
            tsk_vs_close(vs_info);
            return 1;
        }
        tsk_vs_close(vs_info);
    }
    return 0;
}



void process_bloomfile(const char *fname)
{
    if(fname[0]!='+' && fname[0]!='-'){
	fprintf(stderr,"usage: -b[+/-]bloomfile.bloom\n");
	usage();
    }
    bloomset_element::Action action = (fname[0]=='+' ? bloomset_element::INCLUDE : bloomset_element::EXCLUDE);
    bloomset_element *be = new bloomset_element(action, fname+1);
    mybloomset.push_back(be);
}



int main(int argc, char *argv[])
{
    int ch;
    extern int optind;
    const char *arff_fn = 0;
    const char *text_fn = 0;
    const char *xml_fn = 0;

    int t0 = time(0);

    while ((ch = getopt(argc, argv, "mdv1s:A:X:T:n:c:ub:B:SVfC:")) > 0) {
	switch (ch) {
	case 'V': print_version();exit(0);
	case '1': opt_sha1++;break;
	case 'm': opt_md5++; break;
	case 'd': opt_debug++; break;
	case 'v': tsk_verbose++; break; 			// sleuthkit option	  
	case 's': save_outdir = optarg; opt_save = true; break;
	case 'f': opt_magic = 1;break;
	case 'T': text_fn = optarg;break;
	case 'A': arff_fn = optarg;break;
	case 'X': xml_fn = optarg;break;
	case 'n': namelist.push_back(optarg);break;
	case 'c': config_file = optarg; break;
	case 'C': file_count_max = atoi(optarg);break;
	case 'u': opt_userguess = true; break;
	case 'B': save_bloom512 = optarg; break;
	case 'b': process_bloomfile(optarg);break;
	case 'S': opt_multithreaded = 0;break;
	case 'h':
	case '?':
	default:
	    fprintf(stderr, "Invalid argument: %s\n", argv[optind]);
	    usage();
	}
    }

    if (optind >= argc) usage();
    argc -= optind;
    argv += optind;

    const char *filename = argv[0];
    if(!filename){
	errx(1,"must provide filename");
    }


    if(text_fn){
	if(access(text_fn,F_OK)==0) errx(1,"%s: file exists",text_fn);
	t = fopen(text_fn,"w");
	if(!t) err(1,text_fn);
    }

    if(arff_fn){
	if(access(arff_fn,F_OK)==0) errx(1,"%s: file exists",arff_fn);
	FILE *arff_out = fopen(arff_fn,"w");
	if(!arff_out) err(1,arff_fn);
	a = new arff("fiwalk");		// the ARFF output object
	a->set_outfile(arff_out);
    }

    if(xml_fn){
	if(access(xml_fn,F_OK)==0) errx(1,"%s: file exists",xml_fn);
	x = new xml();
	x->setoutfile(xml_fn);
    }

    if(a==0 && x==0 && t==0){		// no default; make default text to stdout
	t = stdout;
    }

    if(config_file) config_read(config_file);    /* Read the configuration file */

    if(strstr(filename,".aff")){
#ifndef HAVE_AFFLIB
	fprintf(stderr,"ERROR: fiwalk was compiled without AFF support.\n");
	exit(0);
#else
	if((tsk_img_type_supported() & TSK_IMG_TYPE_AFF_AFF)==0){
	    fprintf(stderr,"ERROR: fiwalk was compiled with AFF support but the TSK library is not.\n");
	    exit(0);
	}
#endif
    }

    /* If we are outputing ARFF, create the ARFF object and set the file types for a few */
    if(a){
	a->add_attribute("id",arff::NUMERIC);
	a->add_attribute("partition",arff::NUMERIC);
	a->add_attribute("filesize",arff::NUMERIC);
	a->add_attribute("mtime",arff::DATE);
	a->add_attribute("ctime",arff::DATE);
	a->add_attribute("atime",arff::DATE);
	a->add_attribute("fragments",arff::NUMERIC);
	a->add_attribute("frag1startsector",arff::NUMERIC);
	a->add_attribute("frag2startsector",arff::NUMERIC);
	a->add_attribute("filename",arff::STRING);
	if(opt_md5) a->add_attribute("md5",arff::STRING);
	if(opt_sha1) a->add_attribute("sha1",arff::STRING);
	if(opt_userguess) a->add_attribute("userguess",arff::STRING);
    }
    if(x) x->push("fiwalk");

    /* Check that we have a valid file format */
    partition_info("fiwalk_version",fiwalk_version);
    partition_info("Start_time",mytime());

    if (opt_debug) printf("calling tsk_img_open(%s)\n",filename);

#ifdef HAVE_AFFLIB
    /* Get the AFFGID if it exists... */
    AFFILE *af = af_open(filename,0,O_RDONLY);
    if(!af) af_err(1,"af_open(%s)",filename);
    
    comment("AFF version: %s",af_version());
    u_char gid[1024];
    char buf[1024];
    size_t gidlen = sizeof(gid);
    if(af_get_seg(af,AF_IMAGE_GID,0,gid,&gidlen)==0){
	partition_info("image_gid",af_hexbuf(buf,sizeof(buf),gid,gidlen,AF_HEXBUF_NO_SPACES));
	fflush(stdout);			// 
    }
    af_close(af);
#endif    



    TSK_IMG_INFO *img_info;
    img_info = tsk_img_open_sing(argv[0],TSK_IMG_TYPE_DETECT);
    if (!img_info){
	comment("TSK Error (img_open) %s",tsk_error_get());
	comment("=EOF= (img_open)");
	exit(1);
    }
    comment("image filename: %s",filename);

    partition_info("Imagefile",filename);
    if (opt_debug) printf("calling do_dimage()\n");
    
    if (proc_vs(img_info,0)){
	comment("TSK Error (do_dimage) %s",tsk_error_get());
    }
    tsk_img_close(img_info);

#ifdef HAVE_GETRUSAGE
    /* Print usage information */
    struct rusage ru;
    memset(&ru,0,sizeof(ru));
    int t1 = time(0);
    comment("clock: %d",t1-t0);
    if(getrusage(RUSAGE_SELF,&ru)==0){
	partition_info("usage_user_seconds",ru.ru_utime.tv_sec);
	partition_info("usage_system_seconds",ru.ru_stime.tv_sec);
	partition_info("usage_maxrss",ru.ru_maxrss);
	partition_info("usage_reclaims",ru.ru_minflt);
	partition_info("usage_faults",ru.ru_majflt);
	partition_info("usage_swaps",ru.ru_nswap);
	partition_info("usage_inputs",ru.ru_inblock);
	partition_info("usage_outputs",ru.ru_oublock);
	partition_info("stop_time",cstr(mytime()));
    }
#endif
    if(a) a->write();
    if(t) comment("=EOF=");
    if(x) {
	x->pop("fiwalk");
	x->write();
    }
    exit(0);
}


